
;; POPULATION CLASS

(defclass population ()
    ((individuals
      :initform nil
      :initarg :individuals
      :accessor individuals)
     (size
      :initform 0
      :initarg :size
      :accessor size)
     (min-fitness
      :initform nil
      :initarg :min-fitness
      :accessor min-fitness)
     (max-fitness
      :initform nil
      :initarg :max-fitness
      :accessor max-fitness)
     (avg-fitness
      :initform nil
      :initarg :avg-fitness
      :accessor avg-fitness)
     (sum-fitness
      :initform nil
      :initarg :sum-fitness
      :accessor sum-fitness)
     (fn-computed           
      :initform 0
      :initarg :fn-computed
      :accessor fn-computed)
     ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda pripoji populaciu b k populacii a 
(defmethod join-populations (a (b population))
   (cond ((null a)
          (setq a b))
         (t
            (incf (size a) (size b))
            (setf (individuals a) (append (individuals a)
                                     (individuals b)))
            (setf (min-fitness a) nil)
            (setf (max-fitness a) nil)
            (setf (avg-fitness a) nil)
            (setf (sum-fitness a) nil)
            (when (not (equal (fn-computed a) (fn-computed b))) 
               (setf (fn-computed a) 0))
            ))
   a)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vytvori populaciu velkosti'size' "metodou half-and-half"

(defun half-and-half-create-population (&optional (size (size-pop *param*)))
   (let ((pop (make-instance 'population))
         (temp (floor (/ size (- (max-in-depth *param*) 1)))))
      
      (setf (size pop) 0)
      (dotimes (i (- (max-in-depth *param*) 1))
         (dotimes (j temp)
            (cond 
                  ((< j (/ temp 2))
                   (add-individual pop 
                     (full-create-individual (+ i 1)))
                   )
                  (T
                    (add-individual pop
                      (grow-create-individual (+ i 1)))
                    )
                  )
         ))     
      (dotimes (i (mod size (- (max-in-depth *param*) 1)))
            (cond 
                  ((< i (/  (mod size (- (max-in-depth *param*) 1)) 2))
                   (add-individual pop 
                     (full-create-individual (max-in-depth *param*)))
                   )
                  (T
                    (add-individual pop
                      (grow-create-individual (max-in-depth *param*)))
                    )
                  )
         )
      pop
      ))


;;;;;;;;;;;;;;;;;;;
;; Copy population

(defmethod copy-population ((pop population))
   (make-instance 'population
      :size (size pop)
      :fn-computed (fn-computed pop)
      :individuals (mapcar #'(lambda (a)
                                (copy-individual a))
                      (individuals pop))
      :min-fitness (min-fitness pop)
      :max-fitness (max-fitness pop)
      :avg-fitness (avg-fitness pop)
      :sum-fitness (sum-fitness pop)
      ))
      
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Rovnaka ako PUSH ibaze item vlozi na koniec place

(defmacro push-end (item place)
   `(setf ,place (append ,place (list ,item))))


   

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vlozi jedinca ind do populacie pop

(defmethod add-individual ((pop population) (ind individual))
   (incf (size pop))   
   (push-end (copy-individual ind) (individuals pop))
   (rem-stat-fitnesses pop)
   (setf (fn-computed pop) 0))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vymaze nahodneho jedinca populacie

(defmethod rem-rand-ind ((pop population))
   (let ((rand-num (random (size pop)))
          (new-pop (make-instance 'population
                     :fn-computed 0)))
      (dotimes (count (size pop))
         (if (not (equal count rand-num)) 
            (add-individual new-pop (nth count (individuals pop))))
         )
      new-pop
      )
   )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;

(defmethod edit-population ((pop population))
   (let ((new-pop (make-instance 'population)))
      (dolist (ind (individuals pop))
         (add-individual new-pop (edit-individual ind)))
      new-pop))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vyrata "adjusted fitnesses" populacie 

(defmethod compute-adjusted-fitnesses ((pop population))
   (when (not (equal (fn-computed pop) 1))
      (dolist (indiv (individuals pop))
         (evaluate indiv))
      (setf (fn-computed pop) 1)      
      )
   pop
   )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vynuluje "statistic fitnesses" populacie

(defmethod rem-stat-fitnesses ((pop population))
   (setf (min-fitness pop) 0)
   (setf (max-fitness pop) 0)
   (setf (avg-fitness pop) 0)    
   (setf (sum-fitness pop) 0)
   pop
   )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vyrata "normalised fitnesses" populacie 

(defmethod compute-norm-fitnesses ((pop population))
   (let ((sum 0.0))
      
      (when (individuals pop)
         (when (equal (fn-computed pop) 0)
            (compute-adjusted-fitnesses pop))
         (when (equal (fn-computed pop) 1)         
            (setq sum (float (reduce #'+ (mapcar #'(lambda (ind) (adjusted-fitness ind)) (individuals 
                                                                                           pop)))))
            (when (equal sum 0) (setq sum 1))
            (dolist (indiv (individuals pop))
               (setf (norm-fitness indiv) (/ (adjusted-fitness indiv) sum)))
            (setf (fn-computed pop) 2)
            (rem-stat-fitnesses pop))
         )
      )
   )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Metoda vyrata "statistic fitnesses" populacie 
;; vrati zoznam obsahujuci jedinca s najlepsou 
;; a najhorsou vhodnostou
         
(defmethod compute-stat-fitnesses ((pop population))
   (let ((min 0)
         (max 0)
         (avg 0)
         (count 1)
         (worst nil)
         (best nil))

      (when (individuals pop)
         (when (not (equal (fn-computed pop) 2))
            (compute-norm-fitnesses pop))
         (setq best (copy-individual (car (individuals pop))))
         (setq worst (copy-individual (car (individuals pop))))
         (setq min (norm-fitness worst))
         (setq max (norm-fitness best))
         (dolist (ind (cdr (individuals pop)) best)
            (let ((fit (norm-fitness ind)))
               (incf count)
               (when (< fit min)
                  (setq min fit)
                  (setq worst (copy-individual ind)))
               (when (> fit max)
                  (setq max fit)
                  (setq best (copy-individual ind)))))
         (setf (min-fitness pop) min)
         (setf (max-fitness pop) max)
         (setf (sum-fitness pop) 1.0)
         (setf (avg-fitness pop) (float (/ 1 count)))
         (list best worst)
         )
      )
   )

(defmethod get-max-ind ((pop population))
   (first (compute-stat-fitnesses pop)))

(defmethod get-min-ind ((pop population))
   (second (compute-stat-fitnesses pop)))

(defmethod sort-by-raw-fitness ((pop population))
   (let ((new-pop (copy-population pop)))
      (compute-adjusted-fitnesses new-pop)
      (setf (individuals new-pop) (sort (individuals new-pop) #'> :key #'raw-fitness))
      new-pop
        )
   )
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; "fitness proportionate selection"

(defmethod fit-prop-selection ((pop population))
   (let ((offspring nil)
         (rand-num 0)
         (sum 0))
      (compute-stat-fitnesses pop)
      (setf rand-num (random (sum-fitness pop)))
      (block cycle     
        (dolist (ind (individuals pop))
           (when (not (< (incf sum (norm-fitness ind)) rand-num))
              (return-from cycle ind))))
      )
   )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; metoda vykona crossover dvoch jedincov populacie

(defmethod crossover-pop ( (pop population) )
   (let ((par1 (copy-individual (fit-prop-selection pop)))
         (par2 (copy-individual (fit-prop-selection pop))))
      (crossover-ind par1 par2)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
(defmethod decimation-of-population ((pop population))
   (let ((temp (sort-by-raw-fitness pop))
         (temp-pop (make-instance 'population))
         (new-pop (make-instance 'population))
         )
      (dotimes (i (* (size pop) (prop-of-decimation *param*)))
         (add-individual temp-pop (nth i (individuals temp))))
      (dotimes (i (size pop))
         (add-individual new-pop (fit-prop-selection temp-pop)))
      new-pop
        )
   )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *pop* (half-and-half-create-population 400))

(defmethod histogram (hist-stream (pop population) max-fitness)
   (let* ((x-min 350)
          (y-min 50)
          (x-size-win 400)
          (y-size-win 200)
          (x-cell 0)
          (y-cell 0)
          (hits (make-list max-fitness :initial-element 0))
          (max-num 0)
          (i 0))
      (setq hist-stream (open-stream 'bitmap-window *lisp-main-window*
                        :io 
                        :title "Histogram"
                        :window-interior (make-box x-min y-min (+ x-min x-size-win (* 2 20)) 
                                           (+ y-min y-size-win (* 2 20)))
                        :user-scrollable nil 
                        ))
      (compute-norm-fitnesses pop)
      (dolist (ind (individuals pop))
         (incf (nth (- (raw-fitness ind) 1) hits)))
      (setf max-num (+ (apply #'max hits) 1))
;      (setf x-cell (floor (/ x-size-win (+ 1 max-fitness))))
;      (setf y-cell (floor (/ y-size-win max-num)))
      (setf x-cell (/ x-size-win (+ 1 max-fitness)))
      (setf y-cell (/ y-size-win max-num))
      
      (draw-line  hist-stream (make-position 20 (+ y-size-win 20))
        (make-position (+ x-size-win 20) (+ y-size-win 20)))
      (draw-line  hist-stream (make-position 20 (+ y-size-win 20))
       (make-position 20 20))
      (print x-cell)
      (print y-cell)
      (print max-num)
      (setf i 0)
;      (while (< i max-num)
;        (draw-s-line hist-stream  20 (- (+ y-size-win 20) (* (+ i 1) y-cell))
;;          (+ x-size-win 20) (- (+ y-size-win 20) (* (+ i 1) y-cell)))
;        (incf i 10)
;         )
      (dotimes (i max-fitness)
         (draw-square hist-stream (floor (+ 20 (* i x-cell))) (+ y-size-win 20)
           (floor (+ 20 (* (+ i 1) x-cell))) (floor (- (+ y-size-win 20) (* (nth i hits) y-cell))))
         )
      )
   )

